@@ -9,10 +9,10 @@ with open(this_file_dir / "task_prompt.txt") as f:
 
 example: ICLExample = {
     "task_prompt": task_prompt.strip(),
-    "constraints_and_conditions": [],
+    "constraints_and_requirements": [],
 }
 
-example["constraints_and_conditions"] = [
+example["constraints_and_requirements"] = [
     'Finish your letter with the phrase, "In pursuit of a peaceful future."',
     "Add a postscript at the end of your letter",
 ]
