@@ -3,4 +3,4 @@ from typing import TypedDict
 
 class ICLExample(TypedDict):
     task_prompt: str
-    constraints_and_conditions: list[str]
+    constraints_and_requirements: list[str]
