@@ -5,6 +5,7 @@ import typer
 from cli.alora.commands import alora_app
 from cli.decompose import app as decompose_app
 from cli.serve.app import serve
+from cli.eval.commands import eval_app
 
 cli = typer.Typer(name="m", no_args_is_help=True)
 
@@ -12,7 +13,7 @@ cli = typer.Typer(name="m", no_args_is_help=True)
 # Add a default callback for handling the default cli description.
 @cli.callback()
 def callback() -> None:
-    """Perform M Tasks"""
+    """Perform M Tasks."""
 
 
 # Typer assumes that all commands are in the same file/module.
@@ -25,3 +26,5 @@ cli.command(name="serve")(serve)
 # as documented: https://typer.tiangolo.com/tutorial/subcommands/add-typer/#put-them-together.
 cli.add_typer(alora_app)
 cli.add_typer(decompose_app)
+
+cli.add_typer(eval_app)
