@@ -1,6 +1,8 @@
 """Useful type definitions for models, formatters, and backends."""
 
+from enum import Enum
 from typing import Any
+from urllib.parse import urlparse
 
 from mellea.helpers.fancy_logger import FancyLogger
 
@@ -12,12 +14,17 @@ class ModelOption:
 
     Create a dictionary containing model options like this:
 
+    ```python
     from mellea.backends.types import ModelOption
-    model_options = { ModelOption.TEMPERATURE : 0.0, ModelOption.SYSTEM_PROMPT : "You are a helpful assistant" }
+    model_options = {
+        ModelOption.TEMPERATURE : 0.0,
+        ModelOption.SYSTEM_PROMPT : "You are a helpful assistant"
+    }
+    ```
     """
 
     TOOLS = "@@@tools@@@"
-    """Must be a list of callables or a dict[str, Callable]."""
+    """Must be a list[Callable] or a dict[str, Callable] where str is the name of the function."""
 
     MAX_NEW_TOKENS = "@@@max_new_tokens@@@"
     SYSTEM_PROMPT = "@@@system_prompt@@@"
@@ -25,6 +32,7 @@ class ModelOption:
     CONTEXT_WINDOW = "@@@context_window@@@"
     THINKING = "@@@thinking@@@"
     SEED = "@@@seed@@@"
+    STREAM = "@@@stream@@@"
 
     @staticmethod
     def replace_keys(options: dict, from_to: dict[str, str]) -> dict[str, Any]:
@@ -108,3 +116,27 @@ class ModelOption:
             for k, v in overwrite_opts.items():
                 new_options[k] = v
         return new_options
+
+
+class _ServerType(Enum):
+    """Different types of servers that might be relevant for a backend."""
+
+    UNKNOWN = 0
+    LOCALHOST = 1
+    OPENAI = 2
+    REMOTE_VLLM = 3
+    """Must be set manually for now."""
+
+
+def _server_type(url: str) -> _ServerType:
+    """Find a server type based on the url."""
+    try:
+        parsed = urlparse(url)
+        hostname = parsed.hostname
+        if hostname in ("localhost", "127.0.0.1", "::1", "0.0.0.0"):
+            return _ServerType.LOCALHOST
+        elif hostname == "api.openai.com":
+            return _ServerType.OPENAI
+    except Exception as e:
+        print(f"Error parsing URL: {e}")
+    return _ServerType.UNKNOWN
