@@ -1,9 +1,13 @@
 """A generic WatsonX.ai compatible backend that wraps around the watson_machine_learning library."""
 
+import asyncio
 import datetime
+import functools
 import json
 import os
-from collections.abc import Callable
+import warnings
+from collections.abc import AsyncGenerator, Callable, Coroutine
+from dataclasses import fields
 from typing import Any
 
 from ibm_watsonx_ai import APIClient, Credentials
@@ -19,18 +23,30 @@ from mellea.backends.tools import (
     convert_tools_to_json,
 )
 from mellea.backends.types import ModelOption
+from mellea.helpers.async_helpers import (
+    ClientCache,
+    get_current_event_loop,
+    send_to_queue,
+)
 from mellea.helpers.fancy_logger import FancyLogger
+from mellea.helpers.openai_compatible_helpers import (
+    chat_completion_delta_merge,
+    extract_model_tool_requests,
+)
 from mellea.stdlib.base import (
     CBlock,
     Component,
     Context,
     GenerateLog,
+    GenerateType,
     ModelOutputThunk,
     ModelToolCall,
 )
 from mellea.stdlib.chat import Message
 from mellea.stdlib.requirement import ALoraRequirement  # type: ignore
 
+format: None = None  # typing this variable in order to shadow the global format function and ensure mypy checks for errors
+
 
 class WatsonxAIBackend(FormatterBackend):
     """A generic backend class for watsonx SDK."""
@@ -55,7 +71,16 @@ class WatsonxAIBackend(FormatterBackend):
             model_options : Global model options to pass to the model. Defaults to None.
             api_key : watsonx API key. Defaults to None.
             project_id : watsonx project ID. Defaults to None.
+            kwargs : extra kwargs passed to model inference creation.
         """
+        # There are bugs with the Watsonx python sdk related to async event loops;
+        # using the same watsonx backend across multiple event loops causes errors.
+        warnings.warn(
+            "Watsonx Backend is deprecated, use 'LiteLLM' or 'OpenAI' Backends instead",
+            DeprecationWarning,
+            2,
+        )
+
         super().__init__(
             model_id=model_id,
             formatter=(
@@ -71,26 +96,27 @@ class WatsonxAIBackend(FormatterBackend):
             base_url = f"{os.environ.get('WATSONX_URL')}"
         if api_key is None:
             api_key = os.environ.get("WATSONX_API_KEY")
+
         if project_id is None:
             project_id = os.environ.get("WATSONX_PROJECT_ID")
+        self._project_id = project_id
 
-        _creds = Credentials(url=base_url, api_key=api_key)
-        _client = APIClient(credentials=_creds)
-        self._model = ModelInference(
-            model_id=self._get_watsonx_model_id(),
-            api_client=_client,
-            credentials=_creds,
-            project_id=project_id,
-            params=model_options,
-            **kwargs,
-        )
+        self._creds = Credentials(url=base_url, api_key=api_key)
+        self._kwargs = kwargs
+
+        self._client_cache = ClientCache(2)
+
+        # Call once to set up the model inference and prepopulate the cache.
+        _ = self._model
 
         # A mapping of common options for this backend mapped to their Mellea ModelOptions equivalent.
         # These are usually values that must be extracted before hand or that are common among backend providers.
         self.to_mellea_model_opts_map_chats = {
             "system": ModelOption.SYSTEM_PROMPT,
-            "max_tokens": ModelOption.MAX_NEW_TOKENS,
+            "max_tokens": ModelOption.MAX_NEW_TOKENS,  # Is being deprecated in favor of `max_completion_tokens.`
+            "max_completion_tokens": ModelOption.MAX_NEW_TOKENS,
             "tools": ModelOption.TOOLS,
+            "stream": ModelOption.STREAM,
         }
         # A mapping of Mellea specific ModelOptions to the specific names for this backend.
         # These options should almost always be a subset of those specified in the `to_mellea_model_opts_map`.
@@ -98,13 +124,14 @@ class WatsonxAIBackend(FormatterBackend):
         # will be omitted here so that they will be removed when model_options are processed
         # for the call to the model.
         self.from_mellea_model_opts_map_chats = {
-            ModelOption.MAX_NEW_TOKENS: "max_tokens"
+            ModelOption.MAX_NEW_TOKENS: "max_completion_tokens"
         }
 
         # See notes above.
         self.to_mellea_model_opts_map_completions = {
             "random_seed": ModelOption.SEED,
             "max_new_tokens": ModelOption.MAX_NEW_TOKENS,
+            "stream": ModelOption.STREAM,
         }
         # See notes above.
         self.from_mellea_model_opts_map_completions = {
@@ -112,6 +139,25 @@ class WatsonxAIBackend(FormatterBackend):
             ModelOption.MAX_NEW_TOKENS: "max_new_tokens",
         }
 
+    @property
+    def _model(self) -> ModelInference:
+        """Watsonx's client gets tied to a specific event loop. Reset it if needed here."""
+        key = id(get_current_event_loop())
+
+        _model_inference = self._client_cache.get(key)
+        if _model_inference is None:
+            _client = APIClient(credentials=self._creds)
+            _model_inference = ModelInference(
+                model_id=self._get_watsonx_model_id(),
+                api_client=_client,
+                credentials=self._creds,
+                project_id=self._project_id,
+                params=self.model_options,
+                **self._kwargs,
+            )
+            self._client_cache.put(key, _model_inference)
+        return _model_inference
+
     def _get_watsonx_model_id(self) -> str:
         """Gets the watsonx model id from the model_id that was provided in the constructor. Raises AssertionError if the ModelIdentifier does not provide a watsonx_name."""
         watsonx_model_id = (
@@ -126,7 +172,10 @@ class WatsonxAIBackend(FormatterBackend):
 
     def filter_chat_completions_kwargs(self, model_options: dict) -> dict:
         """Filter kwargs to only include valid watsonx chat.completions.create parameters."""
-        chat_params = TextChatParameters.get_sample_params().keys()
+        # TextChatParameters.get_sample_params().keys() can't be completely trusted. It doesn't always contain all
+        # all of the accepted keys. In version 1.3.39, max_tokens was removed even though it's still accepted.
+        # It's a dataclass so use the fields function to get the names.
+        chat_params = {field.name for field in fields(TextChatParameters)}
         return {k: v for k, v in model_options.items() if k in chat_params}
 
     def _simplify_and_merge(
@@ -143,6 +192,7 @@ class WatsonxAIBackend(FormatterBackend):
 
         Args:
             model_options: the model_options for this call
+            is_chat_context: set to True if used for chat completion apis
 
         Returns:
             a new dict
@@ -168,6 +218,7 @@ class WatsonxAIBackend(FormatterBackend):
 
         Args:
             model_options: the model_options for this call
+            is_chat_context: set to True if used for chat completion apis
 
         Returns:
             a new dict
@@ -185,38 +236,36 @@ class WatsonxAIBackend(FormatterBackend):
 
         return model_opts
 
-    def generate_from_context(
+    async def generate_from_context(
         self,
         action: Component | CBlock,
         ctx: Context,
         *,
         format: type[BaseModelSubclass] | None = None,
         model_options: dict | None = None,
-        generate_logs: list[GenerateLog] | None = None,
         tool_calls: bool = False,
     ):
         """See `generate_from_chat_context`."""
         assert ctx.is_chat_context, NotImplementedError(
             "The watsonx.ai backend only supports chat-like contexts."
         )
-        return self.generate_from_chat_context(
+        mot = await self.generate_from_chat_context(
             action,
             ctx,
-            format=format,
+            _format=format,
             model_options=model_options,
-            generate_logs=generate_logs,
             tool_calls=tool_calls,
         )
+        return mot, ctx.add(action).add(mot)
 
-    def generate_from_chat_context(
+    async def generate_from_chat_context(
         self,
         action: Component | CBlock,
         ctx: Context,
         *,
-        format: type[BaseModelSubclass]
+        _format: type[BaseModelSubclass]
         | None = None,  # Type[BaseModelSubclass] is a class object of a subclass of BaseModel
         model_options: dict | None = None,
-        generate_logs: list[GenerateLog] | None = None,
         tool_calls: bool = False,
     ) -> ModelOutputThunk:
         """Generates a new completion from the provided Context using this backend's `Formatter`."""
@@ -224,7 +273,7 @@ class WatsonxAIBackend(FormatterBackend):
             model_options, is_chat_context=ctx.is_chat_context
         )
 
-        linearized_context = ctx.render_for_generation()
+        linearized_context = ctx.view_for_generation()
         assert linearized_context is not None, (
             "Cannot generate from a non-linear context in a FormatterBackend."
         )
@@ -245,12 +294,12 @@ class WatsonxAIBackend(FormatterBackend):
             conversation.append({"role": "system", "content": system_prompt})
         conversation.extend([{"role": m.role, "content": m.content} for m in messages])
 
-        if format is not None:
+        if _format is not None:
             model_opts["response_format"] = {
                 "type": "json_schema",
                 "json_schema": {
-                    "name": format.__name__,
-                    "schema": format.model_json_schema(),
+                    "name": _format.__name__,
+                    "schema": _format.model_json_schema(),
                     "strict": True,
                 },
             }
@@ -260,7 +309,7 @@ class WatsonxAIBackend(FormatterBackend):
         # Append tool call information if applicable.
         tools: dict[str, Callable] = {}
         if tool_calls:
-            if format:
+            if _format:
                 FancyLogger.get_logger().warning(
                     f"tool calling is superseded by format; will not call tools for request: {action}"
                 )
@@ -274,55 +323,171 @@ class WatsonxAIBackend(FormatterBackend):
             FancyLogger.get_logger().info(f"Tools for call: {tools.keys()}")
 
         formatted_tools = convert_tools_to_json(tools)
-        chat_response = self._model.chat(
-            messages=conversation,
-            tools=formatted_tools,
-            tool_choice_option=(
-                "auto" if formatted_tools and len(formatted_tools) > 0 else "none"
-            ),
-            params=self._make_backend_specific_and_remove(
-                model_opts, is_chat_context=ctx.is_chat_context
-            ),
+
+        chat_response: (
+            Coroutine[Any, Any, AsyncGenerator] | Coroutine[Any, Any, dict] | None
+        ) = None
+
+        stream = model_opts.get(ModelOption.STREAM, False)
+        if stream:
+            chat_response = self._model.achat_stream(
+                messages=conversation,
+                tools=formatted_tools,
+                tool_choice_option=(
+                    "auto" if formatted_tools and len(formatted_tools) > 0 else "none"
+                ),
+                params=self._make_backend_specific_and_remove(
+                    model_opts, is_chat_context=ctx.is_chat_context
+                ),
+            )
+        else:
+            chat_response = self._model.achat(
+                messages=conversation,
+                tools=formatted_tools,
+                tool_choice_option=(
+                    "auto" if formatted_tools and len(formatted_tools) > 0 else "none"
+                ),
+                params=self._make_backend_specific_and_remove(
+                    model_opts, is_chat_context=ctx.is_chat_context
+                ),
+            )
+
+        output = ModelOutputThunk(None)
+        output._context = linearized_context
+        output._action = action
+        output._model_options = model_opts
+
+        # Processing functions only pass the ModelOutputThunk (and current chunk of response). Bind the other vars necessary for
+        # each processing step.
+        output._process = self.processing
+        output._post_process = functools.partial(
+            self.post_processing,
+            conversation=conversation,
+            tools=tools,
+            seed=model_opts.get(ModelOption.SEED, None),
+            _format=_format,
         )
 
-        # If a tool is called, there might not be content in the message.
-        response_message = chat_response["choices"][0]["message"].get("content", "")
-        result = ModelOutputThunk(
-            value=response_message,
-            meta={"oai_chat_response": chat_response["choices"][0]},
-            tool_calls=self._extract_model_tool_requests(tools, chat_response),
+        try:
+            # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
+            # We can also support synchronous calls by adding a flag and changing this ._generate function.
+
+            # This function should always be called from a running event loop so we don't have to worry about
+            # scheduling the task to a specific event loop here.
+            output._generate = asyncio.create_task(
+                send_to_queue(chat_response, output._async_queue)
+            )
+            output._generate_type = GenerateType.ASYNC
+        except RuntimeError as e:
+            # Most likely cause is running this function without an event loop present
+            raise e
+
+        return output
+
+    async def processing(self, mot: ModelOutputThunk, chunk: dict):
+        """Called during generation to add information from a single ChatCompletion or ChatCompletionChunk to the ModelOutputThunk.
+
+        For OpenAI-like APIs, tool call parsing is handled in the post processing step.
+        """
+        if mot._thinking is None:
+            mot._thinking = ""
+        if mot._underlying_value is None:
+            mot._underlying_value = ""
+
+        if len(chunk["choices"]) < 1:
+            return  # Empty chunk. Note: this has some metadata information, but ignoring for now.
+
+        # Watsonx returns dicts. Distinguish streaming and non-streaming based on their fields.
+        not_streaming = chunk["choices"][0].get("message", None) is not None
+        if not_streaming:
+            message: dict = chunk["choices"][0].get("message", dict())
+
+            thinking_chunk = message.get("reasoning_content", None)
+            if thinking_chunk is not None:
+                mot._thinking += thinking_chunk
+
+            content_chunk = message.get("content", "")
+            if content_chunk is not None:
+                mot._underlying_value += content_chunk
+
+            mot._meta["oai_chat_response"] = chunk["choices"][0]
+
+        else:  # Streaming.
+            message_delta: dict = chunk["choices"][0].get("delta", dict())
+
+            thinking_chunk = message_delta.get("reasoning_content", None)
+            if thinking_chunk is not None:
+                mot._thinking += thinking_chunk
+
+            content_chunk = message_delta.get("content", None)
+            if content_chunk is not None:
+                mot._underlying_value += content_chunk
+
+            if mot._meta.get("oai_chat_response_streamed", None) is None:
+                mot._meta["oai_chat_response_streamed"] = []
+            mot._meta["oai_chat_response_streamed"].append(chunk["choices"][0])
+
+    async def post_processing(
+        self,
+        mot: ModelOutputThunk,
+        conversation: list[dict],
+        tools: dict[str, Callable],
+        seed,
+        _format,
+    ):
+        """Called when generation is done."""
+        # Reconstruct the chat_response from chunks if streamed.
+        streamed_chunks = mot._meta.get("oai_chat_response_streamed", None)
+        if streamed_chunks is not None:
+            mot._meta["oai_chat_response"] = chat_completion_delta_merge(
+                streamed_chunks
+            )
+
+        assert mot._action is not None, (
+            "ModelOutputThunks should have their action assigned during generation"
+        )
+        assert mot._model_options is not None, (
+            "ModelOutputThunks should have their model_opts assigned during generation"
         )
 
-        parsed_result = self.formatter.parse(source_component=action, result=result)
+        # OpenAI streamed responses give you chunks of tool calls.
+        # As a result, we have to store data between calls and only then
+        # check for complete tool calls in the post_processing step.
+        tool_chunk = extract_model_tool_requests(tools, mot._meta["oai_chat_response"])
+        if tool_chunk is not None:
+            if mot.tool_calls is None:
+                mot.tool_calls = {}
+            # Merge the tool_chunk dict.
+            for key, val in tool_chunk.items():
+                mot.tool_calls[key] = val
 
-        if generate_logs is not None:
-            assert isinstance(generate_logs, list)
-            generate_log = GenerateLog()
-            generate_log.prompt = conversation
-            generate_log.backend = f"watsonx::{self.model_id!s}"
-            generate_log.model_options = model_opts
-            generate_log.date = datetime.datetime.now()
-            generate_log.model_output = chat_response
-            generate_log.extra = {
-                "format": format,
-                # "thinking": thinking,
-                "tools_available": tools,
-                "tools_called": result.tool_calls,
-                "seed": model_opts.get(ModelOption.SEED, None),
-            }
-            generate_log.result = parsed_result
-            generate_log.action = action
-            generate_logs.append(generate_log)
+        self.formatter.parse(mot._action, mot)
 
-        return parsed_result
+        # Generate the log for this ModelOutputThunk.
+        generate_log = GenerateLog()
+        generate_log.prompt = conversation
+        generate_log.backend = f"watsonx::{self.model_id!s}"
+        generate_log.model_options = mot._model_options
+        generate_log.date = datetime.datetime.now()
+        generate_log.model_output = mot._meta["oai_chat_response"]
+        generate_log.extra = {
+            "format": _format,
+            "tools_available": tools,
+            "tools_called": mot.tool_calls,
+            "seed": seed,
+        }
+        generate_log.result = mot
+        generate_log.action = mot._action
+        mot._generate_log = generate_log
 
-    def _generate_from_raw(
+    async def generate_from_raw(
         self,
         actions: list[Component | CBlock],
+        ctx: Context,
         *,
         format: type[BaseModelSubclass] | None = None,
         model_options: dict | None = None,
-        generate_logs: list[GenerateLog] | None = None,
+        tool_calls: bool = False,
     ) -> list[ModelOutputThunk]:
         """Generates a completion text. Gives the input provided to the model without templating."""
         if format is not None:
@@ -334,42 +499,49 @@ class WatsonxAIBackend(FormatterBackend):
 
         prompts = [self.formatter.print(action) for action in actions]
 
-        responses = self._model.generate(
+        responses = await asyncio.to_thread(
+            self._model.generate,
             prompt=prompts,
             params=self._make_backend_specific_and_remove(
                 model_opts, is_chat_context=False
             ),
         )
 
-        results = [
-            ModelOutputThunk(
-                value=response["results"][0]["generated_text"],
-                meta={"oai_completion_response": response["results"][0]},
+        results = []
+        date = datetime.datetime.now()
+
+        for i, response in enumerate(responses):
+            output = response["results"][0]
+            result = ModelOutputThunk(
+                value=output["generated_text"],
+                meta={
+                    "oai_completion_response": response["results"][0],
+                    "usage": {
+                        "prompt_tokens": output.get("input_token_count", 0),
+                        "completion_tokens": output.get("generated_token_count", 0),
+                        "total_tokens": output.get("input_token_count", 0)
+                        + output.get("generated_token_count", 0),
+                    },
+                },
             )
-            for response in responses
-        ]
 
-        for i, result in enumerate(results):
             self.formatter.parse(actions[i], result)
 
-        if generate_logs is not None:
-            assert isinstance(generate_logs, list)
-            date = datetime.datetime.now()
+            generate_log = GenerateLog()
+            generate_log.prompt = prompts[i]
+            generate_log.backend = f"watsonx::{self.model_id!s}"
+            generate_log.model_options = model_opts
+            generate_log.date = date
+            generate_log.model_output = responses
+            generate_log.extra = {
+                "format": format,
+                "seed": model_opts.get(ModelOption.SEED, None),
+            }
+            generate_log.action = actions[i]
 
-            for i in range(len(prompts)):
-                generate_log = GenerateLog()
-                generate_log.prompt = prompts[i]
-                generate_log.backend = f"watsonx::{self.model_id!s}"
-                generate_log.model_options = model_opts
-                generate_log.date = date
-                generate_log.model_output = responses
-                generate_log.extra = {
-                    "format": format,
-                    "seed": model_opts.get(ModelOption.SEED, None),
-                }
-                generate_log.action = actions[i]
-                generate_log.result = results[i]
-                generate_logs.append(generate_log)
+            result._generate_log = generate_log
+
+            results.append(result)
 
         return results
 
