@@ -9,14 +9,19 @@ from mellea.stdlib.requirement import Requirement
 
 
 def as_markdown_list(ctx: Context) -> list[str] | None:
+    """Attempts to format the last_output of the given context as a markdown list."""
     xs = list()
     raw_output = ctx.last_output()
     assert raw_output is not None
     try:
+        assert raw_output.value is not None
         parsed = mistletoe.Document(raw_output.value)
-        for child in parsed.children:
+        assert parsed.children is not None
+        children = list(parsed.children)
+        for child in children:
             if type(child) is not mistletoe.block_token.List:
                 return None
+        assert child.children is not None
         for item in child.children:
             xs.append(mistletoe.base_renderer.BaseRenderer().render(item))
         return xs
@@ -43,10 +48,13 @@ def _md_table(ctx: Context):
     raw_output = ctx.last_output()
     assert raw_output is not None
     try:
+        assert raw_output.value is not None
         parsed = mistletoe.Document(raw_output.value)
-        if len(parsed.children) != 1:
+        assert parsed.children is not None
+        children = list(parsed.children)
+        if len(children) != 1:
             return False
-        return type(parsed.children[0]) is mistletoe.block_token.Table
+        return type(children[0]) is mistletoe.block_token.Table
     except Exception:
         return False
 
