@@ -1,22 +1,23 @@
 """Requirements are a special type of Component used as input to the "validate" step in Instruct/Validate/Repair design patterns."""
 
 import inspect
+import json
 import re
 from collections.abc import Callable
 from copy import copy
 from typing import Any, overload
 
-from mellea.backends import (
-    Backend,
-    BaseModelSubclass,
+from mellea.backends import Backend, BaseModelSubclass
+from mellea.backends.adapters.adapter import AdapterType
+from mellea.helpers.fancy_logger import FancyLogger
+from mellea.stdlib.base import (
     CBlock,
     Component,
     Context,
     ModelOutputThunk,
+    TemplateRepresentation,
 )
-from mellea.backends.aloras import Alora
-from mellea.helpers.fancy_logger import FancyLogger
-from mellea.stdlib.base import GenerateLog, ModelOutputThunk, TemplateRepresentation
+from mellea.stdlib.intrinsics.intrinsic import Intrinsic
 
 
 def default_output_to_bool(x: CBlock | str) -> bool:
@@ -41,7 +42,13 @@ class ValidationResult:
     """ValidationResults store the output of a Requirement's validation. They can be used to return additional info from validation functions, which is useful for sampling/repairing."""
 
     def __init__(
-        self, result: bool, *, reason: str | None = None, score: float | None = None
+        self,
+        result: bool,
+        *,
+        reason: str | None = None,
+        score: float | None = None,
+        thunk: ModelOutputThunk | None = None,
+        context: Context | None = None,
     ):
         """The result of a requirement's validation.
 
@@ -51,24 +58,41 @@ class ValidationResult:
             result: a boolean that is true if the requirement passed
             reason: a reason for the result
             score: if your validator gives you a score back, you can add this as metadata
+            thunk: if your validator utilizes a backend to generate a response, the ModelOutputThunk returned from that request
+            context: if your validator utilizes a backend to generate a response, the context associated with that response
         """
         self._result = result
         self._reason = reason
         self._score = score
+        self._thunk = thunk
+        self._context = context
 
     @property
     def reason(self) -> str | None:
+        """Reason for the validation result."""
         return self._reason
 
     @property
     def score(self) -> float | None:
+        """An optional score for the validation result."""
         return self._score
 
+    @property
+    def thunk(self) -> ModelOutputThunk | None:
+        """The ModelOutputThunk associated with the validation func if an llm was used to generate the final result."""
+        return self._thunk
+
+    @property
+    def context(self) -> Context | None:
+        """The context associated with validation if a backend was used to generate the final result."""
+        return self._context
+
     def as_bool(self) -> bool:
-        """"""
+        """Return a boolean value based on the result."""
         return self._result
 
     def __bool__(self) -> bool:
+        """Return a boolean value based on the result."""
         return self.as_bool()
 
 
@@ -101,14 +125,13 @@ class Requirement(Component):
         # Used for validation. Do not manually populate.
         self._output: str | None = None
 
-    def validate(
+    async def validate(
         self,
         backend: Backend,
         ctx: Context,
         *,
         format: type[BaseModelSubclass] | None = None,
         model_options: dict | None = None,
-        generate_logs: list[GenerateLog] | None = None,
     ) -> ValidationResult:
         """Chooses the appropriate validation strategy and applies that strategy."""
         if self.validation_fn is not None:
@@ -126,16 +149,16 @@ class Requirement(Component):
             # and its template gets populated with the output correctly.
             req_copy = copy(self)
             req_copy._output = last_output.value
-            llm_as_a_judge_result = backend.generate_from_context(
-                req_copy,
-                ctx,
-                format=format,
-                model_options=model_options,
-                generate_logs=generate_logs,
+            llm_as_a_judge_result, val_ctx = await backend.generate_from_context(
+                req_copy, ctx, format=format, model_options=model_options
             )
+            await llm_as_a_judge_result.avalue()
+
             return ValidationResult(
                 result=self.output_to_bool(llm_as_a_judge_result),
                 reason=llm_as_a_judge_result.value,
+                thunk=llm_as_a_judge_result,
+                context=val_ctx,
             )
 
     def parts(self):
@@ -163,19 +186,129 @@ class LLMaJRequirement(Requirement):
     use_aloras: bool = False
 
 
-class ALoraRequirement(Requirement):
+def requirement_check_to_bool(x: CBlock | str) -> bool:
+    """Checks if a given output should be marked converted to `True`.
+
+    By default, the requirement check alora outputs: `{"requirement_likelihood": 0.0}`.
+    True if >.5
+    """
+    output = str(x)
+    req_dict: dict[str, Any] = json.loads(output)
+
+    likelihood = req_dict.get("requirement_likelihood", None)
+    if likelihood is None:
+        FancyLogger.get_logger().warning(
+            f"could not get value from alora requirement output; looking for `requirement_likelihood` in {req_dict}"
+        )
+        return False
+
+    if likelihood > 0.5:
+        return True
+
+    return False
+
+
+class ALoraRequirement(Requirement, Intrinsic):
     """A requirement that always uses an (possibly specified) ALora. If an exception is thrown during the ALora execution path, `mellea` will fall back to LLMaJ. But that is the only case where LLMaJ will be used."""
 
-    def __init__(self, description: str, alora: Alora | None = None):
+    def __init__(self, description: str, intrinsic_name: str | None = None):
         """A requirement that is validated by an ALora.
 
         Args:
             description: See `Requirement.__init__`
-            alora: if None, the ALora with name "constraint" will be used.
+            intrinsic_name: the name of the intrinsic; must match the adapter
         """
-        super().__init__(description, validation_fn=None)
+        # TODO: We may want to actually do the validation_fn here so that we can set the score.
+        super().__init__(
+            description, validation_fn=None, output_to_bool=requirement_check_to_bool
+        )
         self.use_aloras: bool = True
-        self.alora = alora
+
+        if intrinsic_name is None:
+            intrinsic_name = "requirement_check"
+
+        # Initialize the other side of the inheritance tree
+        Intrinsic.__init__(
+            self,
+            intrinsic_name=intrinsic_name,
+            intrinsic_kwargs={"requirement": f"{self.description}"},
+        )
+
+
+class ScorerRequirement(Requirement):
+    """A requirement that always returns a non-None score. The scorer must also define a preference ordering to indicate whether the goal is to maximize or minimize the score."""
+
+    def __init__(
+        self,
+        description: str | None = None,
+        validation_fn: Callable[[Context], ValidationResult] | None = None,
+        preference_ordering: str = "max",
+        *,
+        output_to_bool: Callable[[CBlock | str], bool] | None = default_output_to_bool,
+        check_only: bool = False,
+    ):
+        """A requirement that is validated by an ALora.
+
+        Args:
+            description: See `Requirement.__init__`
+            validation_fn:  If provided, this function will be executed instead of using LLM-as-a-Judge. This function must return a valid score
+            preference_ordering: indicates whether the goal is to maximize or minimize the score. must be either "max" or "min". Defaults to None
+            output_to_bool: See `Requirement.__init__`
+            check_only: See `Requirement.__init__`
+        """
+        super().__init__(
+            description,
+            validation_fn=validation_fn,
+            output_to_bool=output_to_bool,
+            check_only=check_only,
+        )
+
+        if preference_ordering.lower() not in ["max", "min"]:
+            raise NotImplementedError
+        self.preference_ordering: str = preference_ordering.lower()
+
+    async def validate(
+        self,
+        backend: Backend,
+        ctx: Context,
+        *,
+        format: type[BaseModelSubclass] | None = None,
+        model_options: dict | None = None,
+    ) -> ValidationResult:
+        """Chooses the appropriate validation strategy and applies that strategy. Asserts that the returned ValidationResult has a valid score."""
+        if self.validation_fn is not None:
+            # Python validation strategy
+            validation_result = self.validation_fn(ctx)
+            assert validation_result._score is not None, (
+                "ScorerRequirement must have a score that is not None"
+            )
+            return validation_result
+        else:
+            # LLMaJ validation strategy. This includes ALora because the backend generate call will appropriately dispatch.
+            # For ScorerRequirement, provide score of 1 for result=True, 0 for result=False
+            assert self.output_to_bool is not None
+            last_output = ctx.last_output()
+            assert isinstance(last_output, ModelOutputThunk), (
+                " Context has no appropriate last output"
+            )
+
+            # Create a copy of the requirement that holds the output
+            # and its template gets populated with the output correctly.
+            req_copy = copy(self)
+            req_copy._output = last_output.value
+            llm_as_a_judge_result, val_ctx = await backend.generate_from_context(
+                req_copy, ctx, format=format, model_options=model_options
+            )
+            await llm_as_a_judge_result.avalue()
+            result = self.output_to_bool(llm_as_a_judge_result)
+
+            return ValidationResult(
+                result=result,
+                reason=llm_as_a_judge_result.value,
+                score=1 if result else 0,
+                thunk=llm_as_a_judge_result,
+                context=val_ctx,
+            )
 
 
 def reqify(r: str | Requirement) -> Requirement:
